#include <pthread.h>
#include <unistd.h> /* Unix specific include file */
#include <stdlib.h>
#include <stdio.h>

#define THREAD_NR	16

static volatile int cont = 1;

struct list {
	struct list *next;
	struct list *prev;
};

static void list_init(struct list *list)
{
	list->next = list;
	list->prev = list;
}

static void list_insert_between(struct list *new, struct list *prev,
			struct list *next)
{
	new->prev = prev;
	new->next = next;
	prev->next = new;
	next->prev = new;
}

static void list_insert_after(struct list *new, struct list *prev)
{
	list_insert_between(new, prev, prev->next);
}

static void list_remove(struct list *node)
{
	struct list *prev = node->prev;
	struct list *next = node->next;

	prev->next = next;
	next->prev = prev;
}

struct set_node {
	struct list ll;
	int value;
};

struct set {
	pthread_rwlock_t lock;
	struct list list;
};

static struct list *set_lookup_pos(struct set *set, int value)
{
	struct list *head = &set->list;
	struct list *ptr = head->next;

	while (ptr != head) {
		struct set_node *node = (struct set_node *)ptr;

		if (node->value > value)
			break;
		ptr = ptr->next;
	}
	return ptr->prev;
}

static void set_setup(struct set *set)
{
	list_init(&set->list);
	pthread_rwlock_init(&set->lock, NULL);
}

static void set_release(struct set *set)
{
	struct list *head = &set->list;
	struct list *ptr = head->next;

	pthread_rwlock_destroy(&set->lock);
	while (ptr != head) {
		struct list_node *node = (struct list_node *)ptr;

		ptr = ptr->next;
		free(node);
	}
}

static void set_add(struct set *set, int value)
{
	struct set_node *node = malloc(sizeof(*node));
	struct list *found;

	pthread_rwlock_wrlock(&set->lock);
	found = set_lookup_pos(set, value);
	list_insert_after(&node->ll, found);
	pthread_rwlock_unlock(&set->lock);
}

static int set_del(struct set *set, int value)
{
	struct list *pos;
	int found = 0;

	pthread_rwlock_wrlock(&set->lock);
	pos = set_lookup_pos(set, value);
	if (pos != &set->list) {
		if ((found = ((struct set_node *)pos)->value == value))
			list_remove(pos);
	}
	pthread_rwlock_unlock(&set->lock);

	if (found)
		free((struct set_node *)pos);
	return found;
}

static int set_lookup(struct set *set, int value)
{
	struct list *pos;
	int found = 0;

	pthread_rwlock_rdlock(&set->lock);
	pos = set_lookup_pos(set, value);
	if (pos != &set->list)
		found = ((struct set_node *)pos)->value == value;
	pthread_rwlock_unlock(&set->lock);

	return found;
}

static void *adder(void *data)
{
	struct set *set = data;

	while (cont) {
		const int value = rand() % 4;

		set_add(set, value);
		printf("value %d added\n", value);
		sleep(1); /* declared in unistd.h */
	}

	return NULL;
}

static void *deleter(void *data)
{
	struct set *set = data;

	while (cont) {
		const int value = rand() % 4;

		if (set_del(set, value))
			printf("value %d deleted\n", value);
		sleep(1); /* declared in unistd.h */
	}

	return NULL;
}

static void *lookuper(void *data)
{
	struct set *set = data;

	while (cont) {
		const int value = rand() % 4;

		if (set_lookup(set, value))
			printf("value %d found\n", value);
		sleep(1); /* declared in unistd.h */
	}

	return NULL;
}

int main(void)
{
	pthread_t ths[THREAD_NR];
	struct set set;
	void *(*worker[])(void *) = {
		&adder,
		&lookuper,
		&lookuper,
		&lookuper,
		&lookuper,
		&deleter
	};

	set_setup(&set);
	for (int i = 0; i != THREAD_NR; ++i) {
		const int size = sizeof(worker)/sizeof(worker[0]);
		const int rc = pthread_create(&ths[i], NULL,
					worker[i % size], &set);

		if (rc) {
			perror("pthread_create");
			exit(2);
		}
	}

	sleep(5); /* declared in unistd.h */
	cont = 0;

	for (int i = 0; i != THREAD_NR; ++i)
		pthread_join(ths[i], NULL);
	set_release(&set);

	return 0;
}
